/*
   gpg.c

   This file is a part of Anubis.
   Copyright (C) 2001, 2002 The Anubis Team.

   Anubis is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   Anubis is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   Anubis is released under the GPL with the additional exemption that
   compiling, linking, and/or using OpenSSL is allowed.
*/

#include "headers.h"
#include "extern.h"

#ifdef HAVE_GPG
#include <gpgme.h>

static int gpgme_init(void);
static char *gpg_sign(char *);
static char *gpg_encrypt_to_users(char *);
static char *gpg_encrypt_to_remailer(char *);

#define fail_if_err(a) do { \
		if (a) { \
			anubis_error("GPGME: %s.", \
			gpgme_strerror(a)); \
		} \
	} while(0)

#define GPGME_REQ_VERSION "0.3.0" /* 0.3.0 or later */

static int gpgme_init(void)
{
	GpgmeError err;

	if ((gpgme_check_version(GPGME_REQ_VERSION)) == 0) {
		anubis_error("Install GPGME version "GPGME_REQ_VERSION" or later.");
		return -1;
	}
	if ((err = gpgme_engine_check_version(GPGME_PROTOCOL_OpenPGP))) {
		anubis_error("GPGME: failed. %s.", gpgme_strerror(err));
		return -1;
	}
	#ifdef GDEBUG
	if (output == DEBUG)
		puts(gpgme_get_engine_info());
	#endif /* GDEBUG */
	return 0;
}

static const char *passphrase_cb(void *hook, const char *desc, void **r_hd)
{
	if (!desc)
		return 0;
	return gpg_passphrase;
}

static char *gpg_sign(char *gpg_data)
{
	GpgmeCtx ctx;
	GpgmeError err;
	GpgmeData in, out;
	char buf[100];
	char *signed_data;
	size_t nread;

	if (gpgme_init() == -1)
		return 0;

	signed_data = (char *)xmalloc(strlen(gpg_data) + 1024);
	memset(buf, 0, sizeof(buf));
	fail_if_err(gpgme_new(&ctx));
	gpgme_set_passphrase_cb(ctx, (GpgmePassphraseCb)passphrase_cb, 0);
	gpgme_set_textmode(ctx, 1);
	gpgme_set_armor(ctx, 1);

	fail_if_err(gpgme_data_new_from_mem(&in, gpg_data, strlen(gpg_data), 0));
	fail_if_err(gpgme_data_new(&out));
	fail_if_err(gpgme_op_sign(ctx, in, out, GPGME_SIG_MODE_CLEAR));
	fflush(0);

	if (topt & T_ERROR) {
		gpgme_release(ctx);
		free(signed_data);
		return 0;
	}

	fail_if_err(gpgme_data_rewind(out));
	while (!(err = gpgme_data_read(out, buf, 100, &nread)))
	{
		strncat(signed_data, buf, nread);
		memset(buf, 0, sizeof(buf));
	}
	if (err != GPGME_EOF)
		fail_if_err(err);

	gpgme_data_release(out);
	gpgme_data_rewind(in);
	gpgme_data_release(in);
	gpgme_release(ctx);
	return signed_data;
}

static char *gpg_encrypt_to_users(char *gpg_data)
{
	GpgmeCtx ctx;
	GpgmeError err;
	GpgmeData in, out;
	GpgmeRecipients rset;
	char buf[100];
	char *encrypted_data;
	int gpg_keys_i, gpg_keys_j = 0;
	char gpg_key[100];
	size_t nread;

	if (gpgme_init() == -1)
		return 0;

	encrypted_data = (char *)xmalloc(strlen(gpg_data) + 1024);
	memset(buf, 0, sizeof(buf));
	memset(gpg_key, 0, sizeof(gpg_key));

	fail_if_err(gpgme_new(&ctx));
	gpgme_set_armor(ctx, 1);
	fail_if_err(gpgme_data_new_from_mem(&in, gpg_data, strlen(gpg_data), 0));
	fail_if_err(gpgme_data_new(&out));
	fail_if_err(gpgme_recipients_new(&rset));

	strncat(gpg_keys, ",", 1);
	for (gpg_keys_i = 0; gpg_keys_i <= strlen(gpg_keys); gpg_keys_i++)
	{
		if (gpg_keys[gpg_keys_i] == ',') { /* comma found, so add KEY-ID */
			fail_if_err(gpgme_recipients_add_name_with_validity(rset, gpg_key,
			GPGME_VALIDITY_FULL));
			memset(gpg_key, 0, sizeof(gpg_key));
			gpg_keys_j = 0;
		}
		else /* it is not a comma, so add char to KEY_ID string */
			gpg_key[gpg_keys_j++] = gpg_keys[gpg_keys_i];
	}
	fail_if_err(gpgme_op_encrypt(ctx, rset, in, out));
	fflush(0);
	fail_if_err(gpgme_data_rewind(out));

	if (topt & T_ERROR) {
		gpgme_release(ctx);
		free(encrypted_data);
		return 0;
	}

	while (!(err = gpgme_data_read(out, buf, 100, &nread)))
	{
		strncat(encrypted_data, buf, nread);
		memset(buf, 0, sizeof(buf));
	}
	if (err != GPGME_EOF)
		fail_if_err(err);

	gpgme_data_release(out);
	gpgme_data_rewind(in);
	gpgme_data_release(in);
	gpgme_release(ctx);
	return encrypted_data;
}

static char *gpg_encrypt_to_remailer(char *gpg_data)
{
	#define EXTRA_GPG_BUF 8192
	GpgmeCtx ctx;
	GpgmeError err;
	GpgmeData in, out;
	GpgmeRecipients rset;
	char buf[100];
	char *data_to_encrypt;
	size_t nread;

	if (gpgme_init() == -1)
		return 0;

	data_to_encrypt = (char *)xmalloc(strlen(gpg_data) + EXTRA_GPG_BUF);
	memset(buf, 0, sizeof(buf));
	fail_if_err(gpgme_new(&ctx));
	gpgme_set_armor(ctx, 1);

	/*
	   Remailer Type-I support (with GPG encryption).
	*/

	if (mopt & M_RMRRT)
		sprintf(data_to_encrypt, "\r\n::\r\nAnon-To: %s"CRLF, rm_rrt);
	else if (mopt & M_RMPOST)
		sprintf(data_to_encrypt, "\r\n::\r\nAnon-Post-To: %s"CRLF, rm_post);
	if ((mopt & M_RMLT) || (mopt & M_RMRLT)) {
		char mailbuf[LINEBUFFER];
		sprintf(mailbuf, "Latent-Time: +%s%s"CRLF,
		rm_lt, (mopt & M_RMRLT) ? "r" : "");
		strncat(data_to_encrypt, mailbuf, strlen(mailbuf));
	}
	strncat(data_to_encrypt, CRLF, 2);
	if (mopt & M_RMHEADER) {
		remcrlf(rm_header);
		strncat(data_to_encrypt, "##"CRLF, 4);
		strncat(data_to_encrypt, rm_header, strlen(rm_header));
		strncat(data_to_encrypt, CRLF, 2);
	}
	strncat(data_to_encrypt, gpg_data, strlen(gpg_data));
	memset(buf, 0, sizeof(buf));

	fail_if_err(gpgme_data_new_from_mem(&in, data_to_encrypt, strlen(data_to_encrypt), 0));
	fail_if_err(gpgme_data_new(&out));
	fail_if_err(gpgme_recipients_new(&rset));
	fail_if_err(gpgme_recipients_add_name_with_validity(rset, gpg_rm_key,
	GPGME_VALIDITY_FULL));
	fail_if_err(gpgme_op_encrypt(ctx, rset, in, out));
	fflush(0);
	fail_if_err(gpgme_data_rewind(out));

	if (topt & T_ERROR) {
		gpgme_release(ctx);
		free(data_to_encrypt);
		return 0;
	}

	memset(data_to_encrypt, 0, strlen(gpg_data) + EXTRA_GPG_BUF);
	while (!(err = gpgme_data_read(out, buf, 100, &nread)))
	{
		strncat(data_to_encrypt, buf, nread);
		memset(buf, 0, sizeof(buf));
	}
	if (err != GPGME_EOF)
		fail_if_err(err);

	gpgme_recipients_release(rset);
	gpgme_data_release(in);
	gpgme_data_release(out);
	gpgme_release(ctx);
	return data_to_encrypt;
}

void check_gpg(void)
{
	char homedir_s[MAXPATHLEN]; /* SUPERVISOR */
	char homedir_c[MAXPATHLEN]; /* CLIENT */

	get_homedir(user_supervisor, homedir_s, sizeof(homedir_s));
	get_homedir(user_client, homedir_c, sizeof(homedir_c));
	setenv("HOME", homedir_c, 1);

	if (mopt & M_GPG_SIGN) { /* Sign a message */
		char *buf = gpg_sign(msgbody);
		if (topt & T_ERROR)
			return;
		xfree(msgbody);
		msgbody = buf;
	}
	if (mopt & M_GPG_ENCRYPT) { /* Encrypt a message body */
		char *buf = gpg_encrypt_to_users(msgbody);
		if (topt & T_ERROR)
			return;
		xfree(msgbody);
		msgbody = buf;
	}
	if (mopt & M_RMGPG) { /* Remailer Type-I */
		char *buf = gpg_encrypt_to_remailer(msgbody);
		if (topt & T_ERROR)
			return;
		xfree(msgbody);
		msgbody = buf;
	}
	setenv("HOME", homedir_s, 1);
	if (gpg_passphrase) {
		memset(gpg_passphrase, 0, strlen(gpg_passphrase));
		xfree(gpg_passphrase);
	}
	return;
}

#endif /* HAVE_GPG */

/* EOF */

