/*
   net.c

   This file is a part of Anubis.
   Copyright (C) 2001, 2002 The Anubis Team.

   Anubis is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   Anubis is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   Anubis is released under the GPL with the additional exemption that
   compiling, linking, and/or using OpenSSL is allowed.
*/

#include "headers.h"
#include "extern.h"

static int connect_directly_to(char *, unsigned int);
static int mread(int, void *, char *);

int make_remote_connection(char *host, unsigned int port)
{
	int sd;
	char host_backup[65];
	unsigned int port_backup;
	memset(host_backup, 0, sizeof(host_backup));

	/*
	   First of all, make a copy of 'host' and 'port'.
	*/

	safe_strcpy(host_backup, host);
	port_backup = port;

	check_all_proxies(host_backup, &port_backup);
	sd = connect_directly_to(host_backup, port_backup);
	if (sd == -1) /* an error */
		return -1;
	if (check_socks_proxy(sd, host, port) == -1)
		return -1;
	if (check_wingates(sd, host, port) == -1)
		return -1;

	return sd;
}

static int connect_directly_to(char *host, unsigned int port)
{
	int sd = 0;
	unsigned long inaddr;
	struct sockaddr_in addr;

	/*
	   Find out the IP address.
	*/

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	info(VERBOSE, "Getting remote host information...");

	inaddr = inet_addr(host);
	if (inaddr != INADDR_NONE)
		memcpy(&addr.sin_addr, &inaddr, sizeof(inaddr));
	else {
		struct hostent *hp = 0;
		hp = gethostbyname(host);
		if (hp == 0) {
			hostname_error(host);
			return -1;
		}
		else {
			if (hp->h_length != 4 && hp->h_length != 8) {
				anubis_error("Illegal address length received for host %s", host);
				return -1;
			}
			else {
				memcpy((char *)&addr.sin_addr.s_addr, hp->h_addr,
				hp->h_length);
			}
		}
	}

	/*
	   Create socket and connect.
	*/

	info(VERBOSE, "Initializing socket...");
	if ((sd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		anubis_error("Can't create stream socket.");
		return -1;
	}
	info(VERBOSE, "Connecting to %s:%u...", host, port);
	if (connect(sd, (struct sockaddr *)&addr, sizeof(addr)) == -1) {
		anubis_error("Couldn't connect to %s:%u. %s.", host, port, strerror(errno));
		return -1;
	}
	else
		info(NORMAL, "Connected to %s:%u", host, port);

	return sd;
}

/*****************
 Bind and listen.
******************/

int bind_and_listen(char *host, unsigned int port)
{
	int sd = 0;
	unsigned long inaddr;
	struct sockaddr_in addr;

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);

	if ((sd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
		anubis_error("Can't create stream socket.");

	if (topt & T_NAMES) {
		inaddr = inet_addr(host);
		if (inaddr != INADDR_NONE)
			memcpy(&addr.sin_addr, &inaddr, sizeof(inaddr));
		else {
			struct hostent *hp = 0;
			hp = gethostbyname(host);
			if (hp == 0)
				hostname_error(host);
			else {
				if (hp->h_length != 4 && hp->h_length != 8)
					anubis_error("Illegal address length received for host %s", host);
				else {
					memcpy((char *)&addr.sin_addr.s_addr, hp->h_addr,
					hp->h_length);
				}
			}
		}
	}
	else
		addr.sin_addr.s_addr = htonl(INADDR_ANY);

	if (bind(sd, (struct sockaddr *)&addr, sizeof(addr)))
		anubis_error("bind() failed: %s.", strerror(errno));
	info(VERBOSE, "Anubis bound to %s:%u", inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));
	if (listen(sd, 5))
		anubis_error("listen() failed: %s.", strerror(errno));
	return sd;
}

/**************
  Send a data
***************/

void swrite(int method, void *sd, char *txt)
{
	int len;

	if (txt == 0)
		return;
	len = strlen(txt);

	if (method == CLIENT) {
		#ifdef HAVE_SSL
		if (topt & T_SSL_CLIENT) {
			if (SSL_write((SSL *)sd, txt, len) <= 0)
				socket_error();
		} else
		#endif /* HAVE_SSL */
		if (!(topt & T_SSL_CLIENT)) {
			if (send((int)sd, txt, len, 0) == -1)
				socket_error();
		}
		#ifdef GDEBUG
		if (output == DEBUG)
			fprintf(stderr, "(%d)SERVER >>> %s", len, txt);
		#endif /* GDEBUG */
	}
	else if (method == SERVER) {
		#ifdef HAVE_SSL
		if (topt & T_SSL_SERVER) {
			if (SSL_write((SSL *)sd, txt, len) <= 0)
				socket_error();
		} else
		#endif /* HAVE_SSL */
		if ((int)sd == -1) { /* stdout */
			write(1, txt, len);
		}
		else if (!(topt & T_SSL_SERVER)) {
			if (send((int)sd, txt, len, 0) == -1)
				socket_error();
		}
		#ifdef GDEBUG
		if (output == DEBUG)
			fprintf(stderr, "(%d)CLIENT >>> %s", len, txt);
		#endif /* GDEBUG */
	}
	return;
}

/**************
  Read a data
***************/

static int mread(int method, void *sd, char *ptr)
{
	static int nread = 0;
	static char *read_ptr = 0;
	static char buf[LINEBUFFER + 1];

	if (nread <= 0) {
	again:
		if (method == CLIENT) {
			#ifdef HAVE_SSL
			if (topt & T_SSL_CLIENT)
				nread = SSL_read((SSL *)sd, buf, LINEBUFFER);
			else
			#endif /* HAVE_SSL */
			if (!(topt & T_SSL_CLIENT))
				nread = recv((int)sd, buf, LINEBUFFER, 0);
		}
		else if (method == SERVER) {
			#ifdef HAVE_SSL
			if (topt & T_SSL_SERVER)
				nread = SSL_read((SSL *)sd, buf, LINEBUFFER);
			else
			#endif /* HAVE_SSL */
			if ((int)sd == -1) { /* stdin */
				nread = read(0, buf, LINEBUFFER - 1);
				remcrlf(buf);
				strncat(buf, CRLF, 2);
			}
			else if (!(topt & T_SSL_SERVER))
				nread = recv((int)sd, buf, LINEBUFFER, 0);
		}
		if (nread < 0) {
			if (errno == EINTR)
				goto again;
			return -1;
		}
		else if (nread == 0)
			return 0;
		read_ptr = buf;
	}
	nread--;
	*ptr = *read_ptr++;
	return 1;
}

int recvline(int method, void *sd, void *vptr, int maxlen)
{
	int n, rc;
	char c, *ptr;

	ptr = vptr;
	for (n = 1; n < maxlen; n++)
	{
		if ((rc = mread(method, sd, &c)) == 1) {
			*ptr++ = c;
			if (c == '\n')
				break;
		}
		else if (rc == 0) {
			if (n == 1)
				return 0;
			else
				break;
		}
		else
			return -1;
	}
	*ptr = 0;
	#ifdef GDEBUG
	if (output == DEBUG) {
		if (method == CLIENT)
			fprintf(stderr, "(%d)SERVER <<< %s", n, (char *)vptr);
		else if (method == SERVER)
			fprintf(stderr, "(%d)CLIENT <<< %s", n, (char *)vptr);
	}
	#endif /* GDEBUG */
	return n;
}

/*****************
  Get a response
******************/

void get_response_smtp(int method, void *sd, char *buf, int size)
{
	int n;
	char line[LINEBUFFER + 1];

	if (buf != 0)
		memset(buf, 0, size);
	do {
		n = recvline(method, sd, line, LINEBUFFER);
		if (buf != 0) {
			strncat(buf, line, size);
			size -= n;
		}
	} while (line[3] == '-');
	return;
}

/**************************
 Close a socket descriptor
***************************/

void close_socket(int sd)
{
	if (sd)
		close(sd);
	return;
}

/* EOF */

