/*
   rc.c

   This file is a part of Anubis.
   Copyright (C) 2001, 2002 The Anubis Team.

   Anubis is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   Anubis is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   Anubis is released under the GPL with the additional exemption that
   compiling, linking, and/or using OpenSSL is allowed.
*/

#include "headers.h"
#include "extern.h"

static void match_options_common(int, char *);
static int  get_regex(int, char *, char *, int);
static char *parse_line_regex(char *);
static void get_action_line(char *);
static void match_action(char *, char *);

#define set(a, b, c) \
	if (strncmp((char *)a, "yes", 3) == 0) { \
		b |= c; \
		return; \
	} \
	else if (strncmp((char *)a, "no", 2) == 0) { \
		b &= ~c; \
		return; \
	} \
	else \
		return;

#define if_empty_quit(a) \
	if (strlen((char *)a) == 0) \
		return;

#define if_empty_set(a, b, c) \
	if (strlen((char *)a) == 0) { \
		b &= ~c; \
		return; \
	} \
	else if (strlen((char *)a) != 0) { \
		b |= c; \
		return; \
	}

void open_rcfile(int method)
{
	int n;
	char homedir[MAXPATHLEN];
	char local_rcfile[] = DEFAULT_LOCAL_RCFILE;
	char *rcfile = 0;
	char *user = 0;

	if (method == SUPERVISOR)
		user = user_supervisor;
	else if (method == CLIENT)
		user = user_client;

	get_homedir(user, homedir, MAXPATHLEN);
	n = strlen(homedir) + strlen(local_rcfile) + 2;
	n = n > MAXPATHLEN ? MAXPATHLEN : n + 1;

	if (method == SUPERVISOR) {
		if (topt & T_ALTRC)
			rcfile = allocbuf(altrc, MAXPATHLEN);
		else {
			if (check_superuser())
				rcfile = allocbuf(DEFAULT_GLOBAL_RCFILE, MAXPATHLEN);
			else {
				rcfile = (char *)xmalloc(n);
				#ifdef HAVE_SNPRINTF
				snprintf(rcfile, n - 1,
				#else
				sprintf(rcfile,
				#endif /* HAVE_SNPRINTF */
				"%s/%s", homedir, local_rcfile);
			}
		}
	}
	else if (method == CLIENT) {
		rcfile = (char *)xmalloc(n);
		#ifdef HAVE_SNPRINTF
		snprintf(rcfile, n - 1,
		#else
		sprintf(rcfile,
		#endif /* HAVE_SNPRINTF */
		"%s/%s", homedir, local_rcfile);
	}

	if (check_filemode(rcfile) == 0) { /* Wrong permissions... */
		free(rcfile);
		return;
	}
	if (method == SUPERVISOR) {
		struct stat st;
		if (stat(rcfile, &st) == 0) {
			if (global_mtime != st.st_mtime)
				global_mtime = st.st_mtime;
			else {
				free(rcfile);
				return;
			}
		}
	}
	fp_rcfile = fopen(rcfile, "r");
	if (fp_rcfile == 0) {
		if (topt & T_ALTRC) {
			anubis_error("Anubis RC file error: %s.", strerror(errno));
			topt &= ~T_ERROR;
		}
	}
	free(rcfile);
	return;
}

void read_rcfile(int method)
{
	char rcline[LINEBUFFER + 1];
	unsigned long rcfile_position = 0;

	if (fp_rcfile == 0)
		return;

	rcfile_position = get_position(BEGIN_CONTROL);
	if (rcfile_position == 0)
		return;

	fseek(fp_rcfile, rcfile_position, SEEK_SET);
	while (fgets(rcline, LINEBUFFER, fp_rcfile) != 0)
	{
		if (strncmp(rcline, "#", 1) == 0 || strncmp(rcline, LF, 1) == 0)
			continue; /* skip: an empty line, comment (#) */
		else if (strncmp(rcline, END_CONFIG, strlen(END_CONFIG)) == 0)
			break; /* THE END */
		else
			match_options_common(method, rcline);
	}
	if (!(topt & T_ALLOW_LOCAL_MTA))
		topt &= ~T_LOCAL_MTA;

	if ((topt & T_ALLOW_EXTERNAL) && (method >= SUPERVISOR)) {
		xfree(extlist);
		rcfile_position = get_position(BEGIN_EXTERNAL);
		if (rcfile_position == 0)
			return;

		fseek(fp_rcfile, rcfile_position, SEEK_SET);
		while (fgets(rcline, LINEBUFFER, fp_rcfile) != 0)
		{
			if (strncmp(rcline, "#", 1) == 0 || strncmp(rcline, LF, 1) == 0)
				continue; /* skip: an empty line, comment (#) */
			else if (strncmp(rcline, END_CONFIG, strlen(END_CONFIG)) == 0)
				break; /* THE END */
			else {
				int size = strlen(rcline);
				if (extlist)
					size += strlen(extlist);
				extlist = (char *)xrealloc((char *)extlist, size + 1);
				strncat(extlist, rcline, size);
			}
		}
	}
	return;
}

void read_rcfile_allsection(void)
{
	char rcline[LINEBUFFER + 1];
	unsigned long rcfile_position = 0;

	if (fp_rcfile == 0)
		return;

	rcfile_position = get_position(BEGIN_ALL);
	if (rcfile_position == 0)
		return;

	fseek(fp_rcfile, rcfile_position, SEEK_SET);
	while (fgets(rcline, LINEBUFFER, fp_rcfile) != 0)
	{
		if (strncmp(rcline, "#", 1) == 0 || strncmp(rcline, LF, 1) == 0)
			continue; /* skip: an empty line, comment (#) */
		else if (strncmp(rcline, END_CONFIG, strlen(END_CONFIG)) == 0)
			break; /* THE END */
		else {
			get_action_line(rcline);
			match_action(rcline, rcline);
		}
	}
	return;
}

void close_rcfile(void)
{
	if (fp_rcfile) {
		if (fclose(fp_rcfile) != 0) {
			anubis_error("Fatal fclose() error. %s.", strerror(errno));
			topt &= ~T_ERROR;
		}
		fp_rcfile = 0;
	}
	return;
}

char *parse_line_option(char *line)
{
	char *ptr = 0;
	ptr = strchr(line, '=');

	if (ptr == 0)
		return line;
	do {
		ptr++;
	} while (*ptr == ' ' || *ptr == '\t');

	remcrlf(ptr);
	return ptr;
}

static void match_options_common(int method, char *rcline)
{
	char *ptr = 0;
	char buf[LINEBUFFER + 1];

	safe_strcpy(buf, rcline);
	change_to_lower(buf);

	if (method == INIT) {
		if (strncmp(buf, "bind", 4) == 0) {
			ptr = parse_line_option(rcline);
			if_empty_quit(ptr);
			parse_hostname_host(ptr, t_hostname, &t_portnumber);
			if (strlen(t_hostname) != 0)
				topt |= T_NAMES;
			return;
		}
	}
	if (method >= SUPERVISOR) {
		if (strncmp(buf, "output", 6) == 0) {
			ptr = parse_line_option(buf);
			if (strncmp(ptr, "silent", 6) == 0)
				output = SILENT;
			else if (strncmp(ptr, "normal", 6) == 0)
				output = NORMAL;
			else if (strncmp(ptr, "verbose", 7) == 0)
				output = VERBOSE;
			#ifdef GDEBUG
			else if (strncmp(ptr, "debug", 5) == 0)
				output = DEBUG;
			#endif /* GDEBUG */
			else /* unknown output mode */
				output = NORMAL;
			return;
		}
		if (strncmp(buf, "allow-local-mta", 15) == 0) {
			set(parse_line_option(buf), topt, T_ALLOW_LOCAL_MTA);
		}
		if (strncmp(buf, "allow-external-processing", 25) == 0) {
			set(parse_line_option(buf), topt, T_ALLOW_EXTERNAL);
		}
		if (strncmp(buf, "user-notprivileged", 18) == 0) {
			safe_strcpy(user_notprivileged, parse_line_option(rcline));
			if_empty_set(user_notprivileged, topt, T_USER_NOTPRIVIL);
		}
	}
	if (method == CLIENT) {
		if (strncmp(buf, "logfile", 7) == 0) {
			ptr = parse_line_option(rcline);
			if_empty_quit(ptr);
			xfree(logfile);
			logfile = allocbuf(ptr, MAXPATHLEN);
			return;
		}
		if (strncmp(buf, "loglevel", 8) == 0) {
			ptr = parse_line_option(buf);
			if (strncmp(ptr, "none", 4) == 0)
				loglevel = NONE;
			else if (strncmp(ptr, "all", 3) == 0)
				loglevel = ALL;
			else /* default: only ERRORS logging */
				loglevel = ERRORS;
			return;
		}
	}
	if (strncmp(buf, "remote-mta", 10) == 0 || strncmp(buf, "server", 6) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		parse_hostname_port(ptr, hostname, &portnumber);
		return;
	}
	if (strncmp(buf, "local-mta", 9) == 0) {
		char *a = 0;
		char *p = 0;
		char tmp[LINEBUFFER + 1];
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);

		a = strchr(ptr, ' '); /* an extra arguments */
		if (a) {
			*a++ = '\0';
			p = strrchr(ptr, '/');
			if (p)
				p++;
			else
				p = ptr;
			#ifdef HAVE_SNPRINTF
			snprintf(tmp, LINEBUFFER,
			#else
			sprintf(tmp,
			#endif /* HAVE_SNPRINTF */
			"%s %s", p, a);
			p = ptr;
			a = tmp;
		}
		else { /* no arguments */
			p = ptr;
			a = strrchr(ptr, '/');
			if (a)
				a++;
			else
				a = ptr;
		}
		xfree(execpath);
		execpath = allocbuf(p, MAXPATHLEN);
		topt |= T_LOCAL_MTA;
		if (topt & T_RCEXECARGS) {
			free_execargs(execargs);
			topt &= ~T_RCEXECARGS;
		}
		execargs = gen_execargs(a);
		topt |= T_RCEXECARGS;
		return;
	}

	/*
	   Proxies.
	*/

	if (strncmp(buf, "socks-proxy", 11) == 0) {
		parse_hostname_port(parse_line_option(rcline), socks_proxy, &socks_port);
		if_empty_set(socks_proxy, topt, T_SOCKS);
	}
	if (strncmp(buf, "socks-v4", 8) == 0) {
		set(parse_line_option(buf), topt, T_SOCKSV4);
	}
	if (strncmp(buf, "socks-username", 14) == 0) {
		safe_strcpy(socks_username, parse_line_option(rcline));
		if_empty_set(socks_username, topt, T_SOCKSUSERNAME);
	}
	if (strncmp(buf, "socks-password", 14) == 0) {
		safe_strcpy(socks_password, parse_line_option(rcline));
		if_empty_set(socks_password, topt, T_SOCKSPASSWORD);
	}
	if (strncmp(buf, "wingate1", 8) == 0) {
		parse_hostname_port(parse_line_option(rcline), wingate1, &wingate1_port);
		if_empty_set(wingate1, topt, T_WG1);
	}
	if (strncmp(buf, "wingate2", 8) == 0) {
		parse_hostname_port(parse_line_option(rcline), wingate2, &wingate2_port);
		if_empty_set(wingate2, topt, T_WG2);
	}

	/*
	   TLS/SSL.
	*/

	#ifdef HAVE_SSL
	if (strncmp(buf, "ssl", 3) == 0) {
		set(parse_line_option(buf), topt, T_SSL);
	}
	if (strncmp(buf, "oneway-ssl", 10) == 0) {
		set(parse_line_option(buf), topt, T_SSL_ONEWAY);
	}
	if (strncmp(buf, "cert", 4) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		xfree(ssl_cert);
		ssl_cert = allocbuf(ptr, MAXPATHLEN);
		if (method == CLIENT)
			topt |= T_SSL_CKCLIENT;
		return;
	}
	if (strncmp(buf, "key", 3) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		xfree(ssl_key);
		ssl_key = allocbuf(ptr, MAXPATHLEN);
		if (method == CLIENT)
			topt |= T_SSL_CKCLIENT;
		return;
	}
	#endif /* HAVE_SSL */

	return;
}

/****************************
 Regular Expressions support
*****************************/

#ifdef HAVE_REGEX
int regex_match(char *regex, char *line)
{
	int ret;
	regex_t preg;

	if ((ret = regcomp(&preg, regex, REG_EXTENDED | REG_NOSUB))) {
		char errbuf[256 + 1];
		regerror(ret, &preg, errbuf, 256);
		anubis_error("regcomp() failed. %s", errbuf);
		topt &= ~T_ERROR;
		return 0;
	}
	ret = regexec(&preg, line, 0, 0, 0);
	regfree(&preg);

	if (ret == REG_NOMATCH)
		return 0;
	else
		return 1;
}
#endif /* HAVE_REGEX */

#ifdef HAVE_PCRE
int regex_match(char *regex, char *line)
{
	pcre *re;
	const char *error;
	int error_offset;
	int rc;

	re = pcre_compile(regex, 0, &error, &error_offset, 0);
	if (re == 0) {
		anubis_error("pcre_compile() failed at offset %d: %s.",
		error_offset, error);
		topt &= ~T_ERROR;
		return 0;
	}
	rc = pcre_exec(re, 0, line, strlen(line), 0, 0, 0, 0);

	if (rc < 0) {
		switch(rc)
		{
			case PCRE_ERROR_NOMATCH:
				break;
			default:
				break;
		}
		return 0;
	}
	return 1;
}
#endif /* HAVE_PCRE */

int read_regex_block(int method, char *regex, int size)
{
	char rcline[LINEBUFFER + 1];

	if (fp_rcfile == 0)
		return 0;

	while (fgets(rcline, LINEBUFFER, fp_rcfile) != 0)
	{
		if (strncmp(rcline, "#", 1) == 0 || strncmp(rcline, LF, 1) == 0)
			continue; /* skip: an empty line, comment (#) */
		else if (strncmp(rcline, END_CONFIG, strlen(END_CONFIG)) == 0)
			break; /* END OF REGEX BLOCK */
		else {
			if (get_regex(method, rcline, regex, size) == 0)
				continue;
			else
				return 1;
		}
	}
	return 0;
}

static int get_regex(int method, char *rcline, char *regex, int size)
{
	char *ptr = 0;
	int rs = 0;

	if (method == HEADER) { /* "IF HEADER =..." */
		rs = regex_match("[ \t]*[Ii][Ff][ \t]+[Hh][Ee][Aa][Dd][Ee][Rr][ \t]*=",
		rcline);
	}
	else if (method == COMMAND) { /* "IF COMMAND =..." */
		rs = regex_match("[ \t]*[Ii][Ff][ \t]+[Cc][Oo][Mm][Mm][Aa][Nn][Dd][ \t]*=",
		rcline);
	}
	if (rs == 0)
		return 0;
	ptr = parse_line_regex(rcline);
	if (strlen(ptr) == 0)
		return 0;
	strncpy(regex, ptr, size);
	return 1;
}

static char *parse_line_regex(char *rcline)
{
	char *ptr = 0;
	ptr = strchr(rcline, '=');
	if (ptr == 0)
		return rcline;
	ptr++;
	remcrlf(ptr);
	return ptr;
}

/**********************
 Read the ACTION-BLOCK
***********************/

int read_action_block(char *line)
{
	char rcline[LINEBUFFER + 1];

	if (fp_rcfile == 0)
		return 0;

	while (fgets(rcline, LINEBUFFER, fp_rcfile) != 0)
	{
		if (strncmp(rcline, "#", 1) == 0 || strncmp(rcline, LF, 1) == 0)
			continue; /* skip: an empty line, comment (#) */
		else if (regex_match("[ \t]*[Ff][Ii][^0-9A-Za-z][ \t]*", rcline))
			break; /* 'fi' - END OF REGEX BLOCK ACTION */
		else {
			get_action_line(rcline);
			match_action(rcline, line);
			return 1;
		}
	}
	return 0;
}

static void get_action_line(char *rcline)
{
	char *ptr = 0;
	char buf[LINEBUFFER + 1];

	safe_strcpy(buf, rcline);
	ptr = buf;
	while (*ptr == ' ' || *ptr == '\t')
		ptr++;

	remcrlf(ptr);
	strncpy(rcline, ptr, LINEBUFFER);
	return;
}

static void match_action(char *rcline, char *line)
{
	char *ptr = 0;
	char buf[LINEBUFFER + 1];

	safe_strcpy(buf, rcline);
	change_to_lower(buf);

	if (strncmp(buf, "add", 3) == 0) {
		int size;
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		size = strlen(ptr) + 3;
		if (addlist)
			size += strlen(addlist);
		addlist = (char *)xrealloc((char *)addlist, size);
		strncat(addlist, ptr, strlen(ptr));
		strncat(addlist, CRLF, 2);
		return;
	}
	if (strncmp(buf, "remove", 6) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		remheaderline(ptr);
		return;
	}
	if (strncmp(buf, "modify", 6) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		strncpy(line, ptr, LINEBUFFER);
		strncat(line, CRLF, 2);
		return;
	}
	if (strncmp(buf, "rot13-subject", 13) == 0) {
		set(parse_line_option(buf), mopt, M_ROT13S);
	}
	if (strncmp(buf, "rot13-body", 10) == 0) {
		set(parse_line_option(buf), mopt, M_ROT13B);
	}
	if (strncmp(buf, "signature-file-attach", 21) == 0) {
		set(parse_line_option(buf), mopt, M_SIGNATURE);
	}
	if (strncmp(buf, "body-clear-attach", 17) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		xfree(body_attach);
		body_attach = allocbuf(ptr, MAXPATHLEN);
		mopt |= M_BODYATTACH;
		mopt |= M_BODYCLEARATTACH;
		return;
	}
	if (strncmp(buf, "body-attach", 11) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		xfree(body_attach);
		body_attach = allocbuf(ptr, MAXPATHLEN);
		mopt |= M_BODYATTACH;
		mopt &= ~M_BODYCLEARATTACH;
		return;
	}
	if (strncmp(buf, "external-body-processor", 23) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		xfree(extpath);
		extpath = allocbuf(ptr, MAXPATHLEN);
		mopt |= M_EXTBODYPROC;
		return;
	}

	/*
	   GnuPG support.
	*/

	#ifdef HAVE_GPG
	if (strncmp(buf, "gpg-passphrase", 14) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		if (gpg_passphrase) {
			memset(gpg_passphrase, 0, strlen(gpg_passphrase));
			xfree(gpg_passphrase);
		}
		gpg_passphrase = allocbuf(ptr, 0);
		mopt |= M_GPG_PASSPHRASE;
		return;
	}
	if (strncmp(buf, "gpg-encrypt", 11) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		xfree(gpg_keys);
		gpg_keys = allocbuf(ptr, 0);
		mopt |= M_GPG_ENCRYPT;
		return;
	}
	if (strncmp(buf, "gpg-sign", 8) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		if (regex_match("[Yy][Ee][Ss]", ptr) == 0
		|| (mopt & M_GPG_PASSPHRASE) == 0) {
			if (gpg_passphrase) {
				memset(gpg_passphrase, 0, strlen(gpg_passphrase));
				xfree(gpg_passphrase);
			}
			gpg_passphrase = allocbuf(ptr, 0);
		}
		mopt |= M_GPG_SIGN;
		return;
	}
	#endif /* HAVE_GPG */

	/*
	   Remailer Type-I support.
	*/

	if (strncmp(buf, "rm-rrt", 6) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		xfree(rm_rrt);
		rm_rrt = allocbuf(ptr, 0);
		mopt |= M_RM;
		mopt |= M_RMRRT;
		return;
	}
	if (strncmp(buf, "rm-post", 7) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		xfree(rm_post);
		rm_post = allocbuf(ptr, 0);
		mopt |= M_RM;
		mopt |= M_RMPOST;
		return;
	}
	#ifdef HAVE_GPG
	if (strncmp(buf, "rm-gpg", 6) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		xfree(gpg_rm_key);
		gpg_rm_key = allocbuf(ptr, 0);
		mopt |= M_RMGPG;
		return;
	}
	#endif /* HAVE_GPG */
	if (strncmp(buf, "rm-header", 9) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		xfree(rm_header);
		rm_header = allocbuf(ptr, 512);
		mopt |= M_RMHEADER;
		return;
	}
	if (strncmp(buf, "rm-lt", 5) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		safe_strcpy(rm_lt, ptr);
		mopt |= M_RMLT;
		return;
	}
	if (strncmp(buf, "rm-rlt", 6) == 0) {
		set(parse_line_option(buf), mopt, M_RMRLT);
	}
	return;
}

/**************************
 Find the 'LINE' position.
***************************/

unsigned long get_position(char *line)
{
	char buf[LINEBUFFER + 1];

	if (fp_rcfile == 0 || line == 0)
		return 0;

	rewind(fp_rcfile);
	while (fgets(buf, LINEBUFFER, fp_rcfile) != 0)
	{
		if (strncmp(buf, line, strlen(line)) == 0)
			return (unsigned long)ftell(fp_rcfile);
	}
	return 0;
}

/* EOF */

