/*
   tunnel.c

   This file is a part of Anubis.
   Copyright (C) 2001, 2002 The Anubis Team.

   Anubis is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   Anubis is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   Anubis is released under the GPL with the additional exemption that
   compiling, linking, and/or using OpenSSL is allowed.
*/

#include "headers.h"
#include "extern.h"

static int  transfer_command(void *, void *, char *);
static void process_command(void *, void *, char *, int);
static void transfer_header(void *, void *);
static void process_header_line(char *);
static void process_header_full(char *);
static void transfer_body(void *, void *);
static void static_body_transfer(void *, void *);
static void dynamic_body_transfer(void *, void *);
static void clear_body_transfer(void *, void *);
static void add_remailer_commands(void *);
static void transform_body(void *sd_server);

/******************
  The Tunnel core
*******************/

void smtp_session(void *sd_client, void *sd_server)
{
	char command[LINEBUFFER + 1];

	/*
	   First of all, transfer a welcome message.
	*/

	info(VERBOSE, "Transferring message(s)...");
	get_response_smtp(CLIENT, sd_server, command, LINEBUFFER);

	if (strncmp(command, "220 ", 4) == 0 && strstr(command, version) == 0) {
		char *ptr = 0;
		char *banner_ptr = 0;
		char host[65];
		char banner_backup[LINEBUFFER + 1];
		safe_strcpy(banner_backup, command);

		if ((banner_ptr = strchr(banner_backup, ' '))) {
			banner_ptr++;
			safe_strcpy(host, banner_ptr);
			if ((ptr = strchr(host, ' ')))
				*ptr = '\0';
			do {
				banner_ptr++;
			} while (*banner_ptr != ' ');
			banner_ptr++;
			#ifdef HAVE_SNPRINTF
			snprintf(command, LINEBUFFER,
			#else
			sprintf(command,
			#endif /* HAVE_SNPRINTF */
			"220 %s (%s) %s", host, version, banner_ptr);
		}
	}
	swrite(SERVER, sd_client, command);

	/*
	   Then process the commands...
	*/

	while (recvline(SERVER, sd_client, command, LINEBUFFER))
	{
		process_command(sd_client, sd_server, command, LINEBUFFER);
		sd_client = remote_client;
		sd_server = remote_server;

		if (topt & T_ERROR)
			break;
		if (strlen(command) == 0)
			continue;

		if (transfer_command(sd_client, sd_server, command) == 0)
			break;
	}
	return;
}

/********************
  THE MAIL COMMANDS
*********************/

static void process_command(void *sd_client, void *sd_server, char *command, int size)
{
	char buf[LINEBUFFER + 1];
	safe_strcpy(buf, command); /* make a back-up */

	if (regex_position) {
		char *ptr = 0;
		char regex1[LINEBUFFER + 1];
		char regex2[LINEBUFFER + 1];
		fseek(fp_rcfile, regex_position, SEEK_SET);
		while (read_regex_block(COMMAND, regex1, LINEBUFFER) != 0)
		{
			if ((ptr = strstr(regex1, "!="))) {
				*ptr = '\0';
				ptr += 2;
				safe_strcpy(regex2, ptr);
			}
			if (regex_match(regex1, command)) { /* TRUE */
				if (ptr) {
					if (regex_match(regex2, command) == 0) { /* FALSE */
						while (read_action_block(command) != 0)
						{ safe_strcpy(buf, command); }
					}
				}
				else {
					while (read_action_block(command) != 0)
					{ safe_strcpy(buf, command); }
				}
			}
		}
	}

	change_to_lower(buf);
	if (strncmp(buf, "starttls", 8) == 0) {

		#ifdef HAVE_SSL
		if (!(topt & T_STARTTLS) || !(topt & T_SSL)) {
			swrite(SERVER, sd_client, "503 5.5.0 TLS not available"CRLF);
			strncpy(command, "", 1);
			return;
		}
		if (topt & T_SSL_FINISHED)
			return;

		/*
		   Make the TLS/SSL connection with ESMTP server.
		*/

		info(NORMAL, "Using the TLS/SSL encryption...");

		if (!(topt & T_LOCAL_MTA)) {
			char reply[LINEBUFFER + 1];
			swrite(CLIENT, sd_server, "starttls"CRLF);
			get_response_smtp(CLIENT, sd_server, reply, LINEBUFFER);

			if (!isdigit((unsigned char)reply[0]) || (unsigned char)reply[0] > '3') {
				remcrlf(reply);
				info(VERBOSE, "WARNING: %s", reply);
				anubis_error("STARTTLS command failed.");
				return;
			}
			ctx_client = init_ssl_client();
			if (topt & T_ERROR)
				return;
			ssl_client = start_ssl_client((int)sd_server, ctx_client);
			if (topt & T_ERROR)
				return;
			sd_server = remote_server = (void *)ssl_client;
		}

		/*
		   Make the TLS/SSL connection with SMTP client
		   (client connected with the Tunnel).
		*/

		if (ssl_cert == 0)
			ssl_cert = allocbuf(DEFAULT_SSL_PEM, MAXPATHLEN);
		if (check_filename(ssl_cert) == 0)
			return;
		if (ssl_key == 0)
			ssl_key = allocbuf(ssl_cert, MAXPATHLEN);
		else {
			if (check_filename(ssl_key) == 0)
				return;
		}

		/*
		   Check file permissions. Ignore if a client hasn't
		   specified a private key or a certificate.
		*/

		if (topt & T_SSL_CKCLIENT) {
			check_filemode(ssl_cert);
			check_filemode(ssl_key);
		}

		ctx_server = init_ssl_server();

		if (topt & T_ERROR) {
			swrite(SERVER, sd_client, "454 4.3.3 TLS not available"CRLF);
			return;
		}
		swrite(SERVER, sd_client, "220 2.0.0 Ready to start TLS"CRLF);
		ssl_server = start_ssl_server((int)sd_client, ctx_server);
		if (topt & T_ERROR)
			return;
		sd_client = remote_client = (void *)ssl_server;
		topt |= T_SSL_FINISHED;

		#else
		swrite(SERVER, sd_client, "503 5.5.0 TLS not available"CRLF);
		#endif /* HAVE_SSL */

		strncpy(command, "", 1);
		return;
	}
	return;
}

static int transfer_command(void *sd_client, void *sd_server, char *command)
{
	char reply[2*LINEBUFFER + 1];
	char buf[LINEBUFFER + 1];

	safe_strcpy(buf, command);
	change_to_lower(buf);
	swrite(CLIENT, sd_server, command);
	if (topt & T_ERROR)
		return 0;

	if (strncmp(buf, "ehlo", 4) == 0) {
		get_response_smtp(CLIENT, sd_server, reply, 2*LINEBUFFER);

		if (strstr(reply, "STARTTLS"))
			topt |= T_STARTTLS; /* Yes, we can use the TLS/SSL encryption. */

		#ifdef HAVE_SSL
		if ((topt & T_SSL) && (topt & T_SSL_ONEWAY) && (topt & T_STARTTLS)
		&& !(topt & T_SSL_FINISHED) && !(topt & T_LOCAL_MTA)) {

			struct sockaddr_in rclient;
			char ehlo[LINEBUFFER + 1];
			#ifdef __socklen_t_defined
			socklen_t addrlen;
			#else
			int addrlen;
			#endif /* __socklen_t_defined */

			/*
			   The 'ONEWAY' method is used when your MUA doesn't
			   support the TLS/SSL, but your MTA does.
			   Make the TLS/SSL connection with ESMTP server.
			*/

			char newreply[LINEBUFFER + 1];
			info(NORMAL, "Using the 'ONEWAY' TLS/SSL encryption...");
			swrite(CLIENT, sd_server, "starttls"CRLF);
			get_response_smtp(CLIENT, sd_server, newreply, LINEBUFFER);

			if (!isdigit((unsigned char)newreply[0]) || (unsigned char)newreply[0] > '3') {
				remcrlf(newreply);
				info(VERBOSE, "WARNING: %s", newreply);
				anubis_error("STARTTLS (ONEWAY) command failed.");
				topt &= ~T_ERROR;
				topt &= ~T_SSL_ONEWAY;
				swrite(SERVER, sd_client, reply);
				return 1;
			}
			ctx_client = init_ssl_client();
			if (topt & T_ERROR) {
				topt &= ~T_ERROR;
				topt &= ~T_SSL_ONEWAY;
				swrite(SERVER, sd_client, reply);
				return 1;
			}
			ssl_client = start_ssl_client((int)sd_server, ctx_client);
			if (topt & T_ERROR) {
				topt &= ~T_ERROR;
				topt &= ~T_SSL_ONEWAY;
				swrite(SERVER, sd_client, reply);
				return 1;
			}
			sd_server = remote_server = (void *)ssl_client;
			topt |= T_SSL_FINISHED;

			/*
			   Send the EHLO command (after the TLS/SSL negotiation).
			*/

			addrlen = sizeof(rclient);
			if (getpeername((int)sd_client, (struct sockaddr *)&rclient, &addrlen) == -1)
				anubis_error("getpeername() failed: %s.", strerror(errno));
			else {
				#ifdef HAVE_SNPRINTF
				snprintf(ehlo, LINEBUFFER,
				#else
				sprintf(ehlo,
				#endif /* HAVE_SNPRINTF */
				"ehlo %s"CRLF,
				(topt & T_ERROR) ? "localhost" : inet_ntoa(rclient.sin_addr));
			}
			topt &= ~T_ERROR;
			swrite(CLIENT, sd_server, ehlo);
			get_response_smtp(CLIENT, sd_server, reply, 2*LINEBUFFER);
		}
		#endif /* HAVE_SSL */

		/*
		   Remove the STARTTLS command from the EHLO list
		   if no SSL is specified, the SSL is not available,
		   or we're using the 'ONEWAY' TLS/SSL encryption.
		*/

		if ((topt & T_STARTTLS) && (!(topt & T_SSL) || (topt & T_SSL_ONEWAY))) {
			char *starttls1 = "250-STARTTLS";
			char *starttls2 = "STARTTLS";
			if (strstr(reply, starttls1))
				remline(reply, starttls1);
			else if (strstr(reply, starttls2))
				remline(reply, starttls2);
		}
	}
	else
		get_response_smtp(CLIENT, sd_server, reply, 2*LINEBUFFER);

	swrite(SERVER, sd_client, reply);
	if (topt & T_ERROR)
		return 0;

	if (isdigit((unsigned char)reply[0]) && (unsigned char)reply[0] < '4') {
		if (strncmp(buf, "quit", 4) == 0)
			return 0; /* The QUIT command */
		else if (strncmp(buf, "rset", 4) == 0) {
			xfree(msgbody);
			xfree(boundary);
			xfree(addlist);
			xfree(remlist);
			mopt = 0;
			topt &= ~T_BOUNDARY;
			topt &= ~T_ERROR;
			if (topt & T_SUPERCLIENT)
				read_rcfile_allsection();
		}
		else if (strncmp(buf, "data", 4) == 0) {
			alarm(1800);
			transfer_header(sd_client, sd_server);
			transfer_body(sd_client, sd_server);
			alarm(0);
			xfree(msgbody);
			xfree(boundary);
			mopt = 0;
			topt &= ~T_BOUNDARY;
			topt &= ~T_ERROR;
			if (topt & T_SUPERCLIENT)
				read_rcfile_allsection();
		}
	}
	return 1; /* OK */
}

/*****************
  MESSAGE HEADER
******************/

static void transfer_header(void *sd_client, void *sd_server)
{
	char *header_buf = 0;
	char header_line[LINEBUFFER + 1];

	header_buf = (char *)xmalloc(1);
	while (recvline(SERVER, sd_client, header_line, LINEBUFFER))
	{
		process_header_line(header_line);
		header_buf = (char *)xrealloc((char *)header_buf,
		strlen(header_buf) + strlen(header_line) + 1);
		strncat(header_buf, header_line, strlen(header_line));

		if (strncmp(header_line, CRLF, 2) == 0) {
			if (remlist) {
				if (strlen(remlist) != 0) {
					char *pos = 0;
					char oneline[LINEBUFFER + 1];
					char pline[LINEBUFFER + 1];
					do
					{
						getoneline(remlist, oneline, LINEBUFFER);
						remcrlf(oneline);
						remline(remlist, oneline);
						pos = header_buf;
						for (;;)
						{
							getoneline(pos, pline, LINEBUFFER);
							remcrlf(pline);
							if (strlen(pline) <= 2)
								break;
							pos = strchr(pos, '\n');
							pos++;
							if (regex_match(oneline, pline))
								remline(header_buf, pline);
						}
					} while (strlen(remlist) != 0);
				}
				xfree(remlist);
			}
			if (addlist) {
				if (strlen(addlist) != 0) {
					remcrlf(header_buf);
					header_buf = (char *)xrealloc((char *)header_buf,
					strlen(header_buf) + strlen(addlist) + 3);
					strncat(header_buf, addlist, strlen(addlist));
					strncat(header_buf, CRLF, 2);
					xfree(addlist);
				}
			}
			process_header_full(header_buf);
			swrite(CLIENT, sd_server, header_buf);
			free(header_buf);
			return;
		}
	}
}

static void process_header_line(char *header_line)
{
	char *p = 0;
	char backup[LINEBUFFER + 1];

	/*
	   The DCM system.
	*/

	p = strstr(header_line, BEGIN_DCM);
	if (p) {
		safe_strcpy(backup, p);
		*p++ = '\r';
		*p++ = '\n';
		*p = '\0';
		p = backup;
		p += dcmlen;
	}
	else
		p = header_line;

	if (regex_position) {
		char *ptr = 0;
		char regex1[LINEBUFFER + 1];
		char regex2[LINEBUFFER + 1];
		fseek(fp_rcfile, regex_position, SEEK_SET);
		while (read_regex_block(HEADER, regex1, LINEBUFFER) != 0)
		{
			if ((ptr = strstr(regex1, "!="))) {
				*ptr = '\0';
				ptr += 2;
				safe_strcpy(regex2, ptr);
			}
			if (regex_match(regex1, p)) { /* TRUE */
				if (ptr) {
					if (regex_match(regex2, p) == 0) { /* FALSE */
						while (read_action_block(header_line) != 0)
						{ }
					}
				}
				else {
					while (read_action_block(header_line) != 0)
					{ }
				}
			}
		}
	}

	/*
	   If there are any attachments, find the BOUNDARY.
	*/

	if (strncmp(header_line, "Content-Type:", 13) == 0) {
		if (mopt & M_BODYCLEARATTACH) {
			remheaderline("^Content-Type:");
			remheaderline("^Content-Transfer-Encoding:");
		}
		else {
			char *ptr1 = 0;
			char *ptr2 = 0;
			char buf[LINEBUFFER + 1];
			safe_strcpy(buf, header_line);
			ptr1 = strstr(buf, "boundary");
			if (ptr1 == 0) {
				ptr1 = strstr(buf, "BOUNDARY");
				if (ptr1 == 0)
					return;
			}
			topt |= T_BOUNDARY;
			ptr2 = parse_line_option(ptr1);
			boundary = (char *)xmalloc(strlen(ptr2) + 3);
			if (*ptr2 == '"') {
				ptr2++;
				sprintf(boundary, "--%s", ptr2);
				ptr2 = strstr(boundary, "\"");
				*ptr2 = '\0';
			}
			else
				sprintf(boundary, "--%s", ptr2);
		}
	}
	return;
}

static void process_header_full(char *header_buf)
{
	if (mopt & M_ROT13S) {
		char *p = 0;
		p = strstr(header_buf, "Subject:");
		if (p) {
			p = strchr(p, ':');
			p++;
			do {
				*p = (islower((unsigned char)*p)
				? 'a'+ (*p - 'a' + 13)%26 : isupper((unsigned char)*p)
				? 'A' + (*p - 'A' + 13)%26 : *p);
				p++;
			} while (*p != '\n');
		}
	}
	return;
}

/***************
  MESSAGE BODY
****************/

static void transfer_body(void *sd_client, void *sd_server)
{
	if (mopt & M_BODYCLEARATTACH)
			clear_body_transfer(sd_client, sd_server);
	else {
		if ((mopt & M_GPG_ENCRYPT) || (mopt & M_GPG_SIGN) || (mopt & M_ROT13B)
		|| (mopt & M_RM) || (mopt & M_SIGNATURE) || (mopt & M_BODYATTACH)
		|| (mopt & M_EXTBODYPROC))
			static_body_transfer(sd_client, sd_server);
		else
			dynamic_body_transfer(sd_client, sd_server);
	}
	return;
}

static void static_body_transfer(void *sd_client, void *sd_server)
{
	int nb = 0;
	char body_line[LINEBUFFER + 1];
	msgbody = (char *)xmalloc(1);

	/*
	   If there are some attachments...
	*/

	if (topt & T_BOUNDARY) {
		nb = strlen(boundary);
		while (recvline(SERVER, sd_client, body_line, LINEBUFFER))
		{
			swrite(CLIENT, sd_server, body_line);
			if (strncmp(body_line, boundary, nb) == 0) {
				while (recvline(SERVER, sd_client, body_line, LINEBUFFER))
				{
					swrite(CLIENT, sd_server, body_line);
					if (strncmp(body_line, CRLF, 2) == 0)
						break;
				}
				break;
			}
		}

		/*
		   Now we have reached the message body...
		*/

		while (recvline(SERVER, sd_client, body_line, LINEBUFFER))
		{
			if (strncmp(body_line, boundary, nb) == 0) {
				remcrlf(msgbody);
				remcrlf(msgbody);
				transform_body(sd_server);
				swrite(CLIENT, sd_server, msgbody);
				#ifdef HAVE_GPG
				if ((mopt & M_GPG_ENCRYPT) || (mopt & M_GPG_SIGN))
					swrite(CLIENT, sd_server, CRLF);
				#endif /* HAVE_GPG */
				swrite(CLIENT, sd_server, body_line);
				break;
			}
			msgbody = (char *)xrealloc((char *)msgbody,
			strlen(msgbody) + strlen(body_line) + 1);
			strncat(msgbody, body_line, strlen(body_line));
		}

		/*
		   Transfer everything else...
		*/

		dynamic_body_transfer(sd_client, sd_server);
	}

	/*
	   else... No attachments.
	*/

	else {
		while (recvline(SERVER, sd_client, body_line, LINEBUFFER))
		{
			if (strncmp(body_line, "."CRLF, 3) == 0) { /* EOM */
				remcrlf(msgbody);
				transform_body(sd_server);
				swrite(CLIENT, sd_server, msgbody);
				swrite(CLIENT, sd_server, "."CRLF);
				recvline(CLIENT, sd_server, body_line, LINEBUFFER);
				swrite(SERVER, sd_client, body_line);
				break;
			}
			msgbody = (char *)xrealloc((char *)msgbody,
			strlen(msgbody) + strlen(body_line) + 1);
			strncat(msgbody, body_line, strlen(body_line));
		}
	}
	return;
}

static void dynamic_body_transfer(void *sd_client, void *sd_server)
{
	char body_line[LINEBUFFER + 1];

	while (recvline(SERVER, sd_client, body_line, LINEBUFFER))
	{
		swrite(CLIENT, sd_server, body_line);
		if (strncmp(body_line, "."CRLF, 3) == 0) { /* EOM */
			recvline(CLIENT, sd_server, body_line, LINEBUFFER);
			swrite(SERVER, sd_client, body_line);
			return;
		}
	}
}

static void clear_body_transfer(void *sd_client, void *sd_server)
{
	char body_line[LINEBUFFER + 1];
	msgbody = (char *)xmalloc(1);

	while (recvline(SERVER, sd_client, body_line, LINEBUFFER))
	{
		if (strncmp(body_line, "."CRLF, 3) == 0) { /* EOM */
			transform_body(sd_server);
			swrite(CLIENT, sd_server, msgbody);
			swrite(CLIENT, sd_server, body_line);
			recvline(CLIENT, sd_server, body_line, LINEBUFFER);
			swrite(SERVER, sd_client, body_line);
			return;
		}
	}
}

static void add_remailer_commands(void *sd_server)
{
	char buf[1024];

	if (!(mopt & M_RMGPG)) {
		if (mopt & M_RMRRT) {
			sprintf(buf, "::\r\nAnon-To: %s"CRLF, rm_rrt);
			swrite(CLIENT, sd_server, buf);
		}
		else if (mopt & M_RMPOST) {
			sprintf(buf, "::\r\nAnon-Post-To: %s"CRLF, rm_post);
			swrite(CLIENT, sd_server, buf);
		}
		if ((mopt & M_RMLT) || (mopt & M_RMRLT)) {
			sprintf(buf, "Latent-Time: +%s%s"CRLF,
			rm_lt, (mopt & M_RMRLT) ? "r" : "");
			swrite(CLIENT, sd_server, buf);
		}
		swrite(CLIENT, sd_server, CRLF);
		if (mopt & M_RMHEADER) {
			remcrlf(rm_header);
			swrite(CLIENT, sd_server, "##"CRLF);
			swrite(CLIENT, sd_server, rm_header);
			swrite(CLIENT, sd_server, CRLF);
		}
	}
	#ifdef HAVE_GPG
	else
		swrite(CLIENT, sd_server, "::\r\nEncrypted: PGP\r\n"CRLF);
	#endif /* HAVE_GPG */

	return;
}

static void transform_body(void *sd_server)
{
	check_all_files(user_client);
	if (!(topt & T_ERROR) && (mopt & M_EXTBODYPROC)) {
		int rs = 0;
		char *extbuf = external_program(&rs, extpath, msgbody, 0, 0);
		if (rs != -1 && extbuf) {
			xfree(msgbody);
			msgbody = extbuf;
		}
	}
	if (!(topt & T_ERROR)) 
		check_rot13();
	#ifdef HAVE_GPG
	if (!(topt & T_ERROR)) 
		check_gpg();
	#endif /* HAVE_GPG */

	if (!(topt & T_ERROR)) {
		if (mopt & M_RM)
			add_remailer_commands(sd_server);
	}
	topt &= ~T_ERROR;
	return;
}

/* EOF */

